// OOBPNET.CPP
//
// This program is an object oriented aritficial neural network model.
// It uses the neuron and synapse classes as defined in the ANN.H file.
// Written by Dariusz Piatkowski, January 8 1996.
//
// USES: ANN.H & ANN.CPP
//
// NOTE: If you use this code in its entirety (complete program) as a part of
// another project please do not remove this notice (i.e. you are free to learn
// from it but please do not copy it without including the above credits).

#include <stdlib.h>
#include <iostream.h>
#include <iomanip.h>
#include <conio.h>
#include <fstream.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include "ann.h"
#include "ann.cpp"

#define TRUE 1
#define FALSE 0

//setup stream types
ifstream train_data_file;
ifstream test_data_file;
ifstream weights_load_data_file;
ofstream weights_save_data_file;
ofstream weights_original_file;
ofstream results_data_file;

//function declarations
int main( int argc, char *argv[]);
void initialize( int argc, char *argv[]);
void setup_file_io(int argc, char *argv[], int &randomize);
void train_net(void);
void run_net(ifstream &file);
void test_net(void);
void reset_net(void);
void backprop_error(void);
void clean_up(int argc);

//declare some global variables
int no_neurons, no_inputs, no_outputs, no_layers, no_s_layers;
int no_hidden[5]={0,0,0,0,0};
long no_syn[6]={0L,0L,0L,0L,0L,0L};
int no_train_data_sets, no_test_data_sets;
long max_epochs, no_synapses;
double input_learn_rate, hidden_learn_rate, steepness, momentum;
double tolerance, desired_net_error;
double net_error=0.0;
neuron *input_n=NULL;
neuron *hidden_n[5]={NULL,NULL,NULL,NULL,NULL};
neuron *output_n=NULL;
synapse *input_s=NULL;
synapse *hidden_s[5]={NULL,NULL,NULL,NULL,NULL};
synapse *output_s=NULL;
long tot_epoch=0L;

int main(int argc, char *argv[])
{
    char answer;

    cout.precision(10);
    initialize(argc, argv);
    if(argc<5)
    {
        train_net();
        test_net();
    }
    else
    {
        cout<<"\nDo you still want to train the net? (y/n): ";
        cin>>answer;
        cin.ignore(1);
        answer=toupper(answer);
        if(answer=='Y')
        {
            train_net();
            test_net();
        }
        else if(answer=='N')
            test_net();
    }
    clean_up(argc);
    return 0;
}

//this function tests the net
void test_net(void)
{
    neuron *temp_n;
    long total=0L, incorrect=0L;
    double success;
    int i, c, good_exampler;

    cout<<"\nTesting the network, please wait...\n";
    for(i=0;i<no_test_data_sets;i++)
    {
        reset_net();
        good_exampler=TRUE;
        run_net(test_data_file);
        temp_n=output_n;
        for(c=0;c<no_outputs;c++)
        {
            results_data_file<<temp_n->desired_output<<" ";
            if(temp_n->real_error>tolerance || temp_n->real_error<(-tolerance))
            {
                good_exampler=FALSE;
                temp_n->good=FALSE;
                incorrect++;
            }
            total++;
            temp_n++;
        }
        results_data_file<<endl;
        temp_n=output_n;
        for(c=0;c<no_outputs;c++)
        {
            if(!temp_n->good)
                results_data_file<<temp_n->output<<" X ";
            else
                results_data_file<<temp_n->output<<" ";
            temp_n++;
        }
        if(!good_exampler)
            results_data_file<<" --> BAD";
        results_data_file<<endl<<endl;
    }
    cout<<"\nThe network output was correct in "<<(total-incorrect)<<" cases,\n";
    cout<<"and incorrect in "<<incorrect<<" cases.\n";
    success=((double)(total-incorrect)/(double)total)*100;
    cout<<"\nThe percentage success rate is: "<<setprecision(3)<<success<<"%.\n";
    cout<<"Training took: "<<tot_epoch<<" epochs.\n";
}

//this function trains the network
void train_net(void)
{
    char inbuf[81];
    long epoch=0L;

    //keep track of how many epochs we've gone through
    do
    {
        net_error=0.0;
        train_data_file.seekg(0);
        train_data_file.getline(inbuf, 81);
        while(!train_data_file.eof())
        {
            run_net(train_data_file); //perform the run-through net calculations
            backprop_error(); //perform network error calculations
            reset_net(); //reset the network values for another run
        }
        //print out some network stats: MSE
        net_error=net_error/(double)(no_train_data_sets*no_outputs);
        epoch++;
        cout<<"Epoch: "<<epoch<<", MSE: "<<net_error<<endl;
        train_data_file.clear();
    }
    while((epoch<max_epochs) && (desired_net_error<=net_error));
    tot_epoch=epoch;
}

//this function performs the run-through net calculations
//including the calculations of output layer real_errors
void run_net(ifstream &file)
{
    neuron *temp_n;
    synapse *temp_s;
    int i, c;
    double exemplar_error=0.0;
    double result;

    //load up all the input values for the input layer neurons
    temp_n=input_n;
    for(i=0;i<no_inputs;i++)
    {
        temp_n->LoadInput(file);
        temp_n++;
    }
    //load up all the output values for the output layer neurons
    temp_n=output_n;
    for(i=0;i<no_outputs;i++)
    {
        temp_n->LoadOutput(file);
        temp_n++;
    }
    //perform the individual layer neuron output calculations
    //then move on to the next layer
    //do all this till all the synapse layers have been processed
    temp_s=input_s;
    temp_n=hidden_n[0];
    for(i=0;i<no_s_layers;i++)
    {
        //multiply output from "from" neuron by synapse weight and assign the
        //value to the input from "to"neuron
        for(c=0;c<no_syn[i];c++)
        {
            temp_s->PassValue();
            temp_s++;
        }
        //perform output (firing) calculations on the neurons of a single layer
        if(i!=no_layers)
            for(c=0;c<no_hidden[i];c++)
            {
                temp_n->Fire();
                temp_n++;
            }
        else
        //if i=no_layers we're calculating the last layer (output layer)
        {
            temp_n=output_n;
            for(c=0;c<no_outputs;c++)
            {
                temp_n->Fire();
                temp_n++;
            }
        }
    }
    //calculate the output layer error
    temp_n=output_n;
    for(c=0;c<no_outputs;c++)
    {
        result=temp_n->CalcError(1); //use the overloaded neuron error function
        exemplar_error+=result;
        net_error+=result;
        temp_n++;
    }
    exemplar_error=exemplar_error/no_outputs;
}

//this function calculates the network errors and backpropagates them
void backprop_error(void)
{
    neuron *temp_n;
    synapse *temp_s;
    int c, i;

    //calculate the hidden layer to output layer errors
    temp_s=output_s;
    for(c=0;c<no_syn[no_layers];c++)
    {
        temp_s->CalcError();
        temp_s++;
    }
    temp_n=hidden_n[no_layers-1];
    for(c=0;c<no_hidden[no_layers-1];c++)
    {
        temp_n->CalcError();
        temp_n++;
    }
    //adjust the connection weight and delta_weight value
    temp_s=output_s;
    for(c=0;c<no_syn[no_layers];c++)
    {
        temp_s->AdjustWeight(hidden_learn_rate);
        temp_s++;
    }
    //process the errors for all of the hidden layers
    if(no_layers>1) //check to see if network has more than 1 hidden layer
    {
        c=no_layers-2;
        while(c>-1)
        {
            temp_s=hidden_s[c];
            for(i=0;i<no_syn[c+1];i++)
            {
                temp_s->CalcError();
                temp_s++;
            }
            temp_n=hidden_n[c];
            for(i=0;i<no_hidden[c];i++)
            {
                temp_n->CalcError();
                temp_n++;
            }
            //adjust the connection weight and delta_weight value
            temp_s=hidden_s[c];
            for(i=0;i<no_syn[c+1];i++)
            {
                temp_s->AdjustWeight(hidden_learn_rate);
                temp_s++;
            }
            c--;
        }
    }
    //adjust the connection weight and delta_weight value
    temp_s=input_s;
    for(c=0;c<no_syn[0];c++)
    {
        temp_s->AdjustWeight(input_learn_rate);
        temp_s++;
    }
    //adjust the output layer neuron's bias
    temp_n=output_n;
    for(c=0;c<no_outputs;c++)
    {
        temp_n->AdjustBias(hidden_learn_rate);
        temp_n++;
    }
    //adjust the hidden layer neuron's bias
    temp_n=hidden_n[0];
    for(c=0;c<(no_neurons-no_inputs-no_outputs);c++)
    {
        temp_n->AdjustBias(input_learn_rate);
        temp_n++;
    }
}

//the following function traverses throughout the whole net and
//resets some values to their default zero values for another run
void reset_net(void)
{
    neuron *temp_n;
    int i;

    temp_n=input_n;
    for(i=0;i<no_neurons;i++)
    {
        temp_n->Reset();
        temp_n++;
    }
}

//the following function initializes the net
void initialize(int argc, char *argv[])
{
    neuron *from_n, *to_n, *temp_n;
    synapse *temp_s,*s;
    int c=0, i, j;
    int load_weights_from_file=FALSE;
    long result=0L;
    time_t t;

    no_synapses=0L;
    no_neurons=0;
    //open the files and read in the network architecture data
    setup_file_io(argc, argv, load_weights_from_file);
    no_s_layers=no_layers+1;
    //calculate the total number of neurons in the network
    while((no_hidden[c]!=0)&&(c<5))
        no_neurons+=no_hidden[c++];
    no_neurons+=no_inputs;
    no_neurons+=no_outputs;
    //display the memory requirements
    cout<<"There are a total of : "<<no_neurons<<" neurons, requiring : ";
    cout<<(sizeof(neuron)*no_neurons)<<" bytes...\n";
    //initialize all the neurons
    if((input_n=new neuron[no_neurons])==NULL)
    {
        cout<<"Not enough memory to run this problem!";
        abort();
    }
    temp_n=input_n;
    temp_n+=no_inputs;
    for(c=0;c<no_layers;c++)
    {
        hidden_n[c]=temp_n;
        temp_n+=no_hidden[c];
    }
    output_n=temp_n;
    //step through all the neurons and reset them to 0.0
    //as well as set the momentum value
    temp_n=input_n;
    for(i=0;i<no_neurons;i++)
    {
        temp_n->Initialize(momentum);
        temp_n++;
    }
    //calculate the total number of synapses in the network
    no_syn[0]=no_inputs*no_hidden[0];
    for(c=0;c<(no_layers-1);c++)
    {
        result=(long)no_hidden[c]*(long)no_hidden[c+1];
        no_syn[c+1]=result;
        no_synapses+=result;
    }
    result=(long)no_hidden[no_layers-1]*(long)no_outputs;
    no_syn[no_layers]=result;
    no_synapses+=result;
    no_synapses+=no_inputs*(long)no_hidden[0];
    //display the memory requirements
    cout<<"\nThere are a total of : "<<no_synapses<<" synapses, requiring : ";
    cout<<(sizeof(synapse)*no_synapses)<<" bytes...\n\n";
    if((input_s=new synapse[no_synapses])==NULL)
    {
        cout<<"Not enough memory to run this problem!";
        abort();
    }
    //set up the major synapse pointers
    temp_s=input_s;
    temp_s+=(no_inputs*no_hidden[0]);
    for(c=0;c<(no_layers-1);c++)
    {
        hidden_s[c]=temp_s;
        temp_s+=no_hidden[c]*no_hidden[c+1];
    }
    hidden_s[c]=temp_s;
    output_s=temp_s;
    //create the network architecture by pointing all the synapses
    //at the right neurons
    //assign all the input layer to hidden layer synapse connections first
    from_n=input_n;
    s=input_s;
    for(c=0;c<no_inputs;c++)
    {
        to_n=hidden_n[0];
        for(i=0;i<no_hidden[0];i++)
        {
            s->from=from_n;
            s->to=to_n;
            to_n++;
            s++;
        }
        from_n++;
    }
    //now assign all the synapses for the hidden layers
    s=hidden_s[0];
    for(c=0;c<(no_layers-1);c++)
    {
        from_n=hidden_n[c];
        for(i=0;i<no_hidden[c];i++)
        {
            to_n=hidden_n[c+1];
            for(j=0;j<no_hidden[c+1];j++)
            {
                s->from=from_n;
                s->to=to_n;
                to_n++;
                s++;
            }
            from_n++;
        }
    }
    //assign all the hidden layer to output layer synapse connections
    from_n=hidden_n[no_layers-1];
    s=output_s;
    for(c=0;c<no_hidden[no_layers-1];c++)
    {
        to_n=output_n;
        for(i=0;i<no_outputs;i++)
        {
            s->from=from_n;
            s->to=to_n;
            to_n++;
            s++;
        }
        from_n++;
    }
    //step through all the synapses and reset the variables to 0.0
    //as well initialize the momentum variable
    temp_s=input_s;
    for(i=0;i<no_synapses;i++)
    {
        temp_s->Initialize(momentum);
        temp_s++;
    }
    //load up the weights and bias if they come from the file
    if(load_weights_from_file)
    {
        temp_s=input_s;
        for(i=0;i<no_synapses;i++)
        {
            weights_load_data_file>>temp_s->weight;
            temp_s++;
        }
        temp_n=input_n;
        for(i=0;i<no_neurons;i++)
        {
            weights_load_data_file>>temp_n->bias;
            temp_n++;
        }
    }
    else
    {
        srand((unsigned)time(&t));  //seed the random # generator
        //assign random weight values to all the synapses
        temp_s=input_s;
        for(i=0;i<no_synapses;i++)
        {
            temp_s->Randomize();
            temp_s++;
        }
        //assign random bias to all the neurons
        temp_n=input_n;
        for(i=0;i<no_neurons;i++)
        {
            temp_n->Randomize();
            temp_n++;
        }
    }
    //write out all the weights, and bias to a file
    temp_s=input_s;
    for(c=0;c<no_synapses;c++)
    {
        weights_original_file<<temp_s->weight<<" ";
        temp_s++;
    }
    weights_original_file<<endl;
    temp_n=input_n;
    for(c=0;c<no_neurons;c++)
    {
        weights_original_file<<temp_n->bias<<" ";
        temp_n++;
    }
}

//this function opens the input and output files
void setup_file_io(int argc, char *argv[], int &load_weights_from_file)
{
    char weights_save_data_filename[12]="";
    int i, flag;

    //open all the data files and check for # of command-line parameters
    if(argc<3)
    {
        cout<<"\nbp train.file test.file (weights.out) (weights.in)\n";
        exit(1);
    }
    if(argc<4)
        strcpy(weights_save_data_filename,"weights.out\0");
    else
        strcpy(weights_save_data_filename, argv[3]);
    //open all the files and check for any errors
    train_data_file.open(argv[1], ios::in);
    test_data_file.open(argv[2], ios::in);
    weights_save_data_file.open(weights_save_data_filename, ios::out);
    weights_original_file.open("weights.org", ios::out);
    results_data_file.open("results", ios::out);
    if(argc==5)
    {
        weights_load_data_file.open(argv[4], ios::in);
        load_weights_from_file=TRUE;
    }
    if(!results_data_file)
    {
        cout<<"\nError occured while opening results file!\n\n";
        exit(1);
    }
    if(!weights_original_file)
    {
        cout<<"\nError occured while opening weights.org file!\n\n";
        exit(1);
    }
    if(!weights_load_data_file)
    {
        cout<<"\nError occured while opening "<<argv[4]<<" file!\n\n";
        exit(1);
    }
    if(!train_data_file)
    {
        cout<<"\nError occured while opening "<<argv[1]<<" file!\n\n";
        exit(1);
    }
    if(!test_data_file)
    {
        cout<<"\nError occured while opening "<<argv[2]<<" file!\n\n";
        exit(1);
    }
    if(!weights_save_data_file)
    {
        cout<<"\nError occured while opening "<<weights_save_data_filename
            <<" file!\n\n";
        exit(1);
    }
    train_data_file.precision(10);
    test_data_file.precision(10);
    weights_load_data_file.precision(10);
    weights_save_data_file.precision(10);
    weights_original_file.precision(10);
    results_data_file.precision(10);
    //input all the info from the train data file
    train_data_file>>no_train_data_sets>>no_test_data_sets>>no_inputs>>
                     no_outputs>>no_layers;
    for(i=0;i<no_layers;i++)
        train_data_file>>no_hidden[i];
    //scan to see if more info on training parameters is in file
    //or whether they are user supplied
    train_data_file>>flag;
    if(flag!=TRUE) //if user supplied get the info
    {
        cout<<"Enter the learning rate for the input layer connections: ";
        cin>>input_learn_rate;
        cin.ignore(1);
        cout<<"\nEnter the learning rate for the hidden layer connections: ";
        cin>>hidden_learn_rate;
        cin.ignore(1);
        cout<<"\nEnter the sigmoid function steepness coefficient: ";
        cin>>steepness;
        cin.ignore(1);
        cout<<"\nEnter the momentum term value: ";
        cin>>momentum;
        cin.ignore(1);
        cout<<"\nEnter the maximum number of epochs to iterate through: ";
        cin>>max_epochs;
        cin.ignore(1);
        cout<<"\nEnter the desired maximum network error: ";
        cin>>desired_net_error;
        cin.ignore(1);
        cout<<"\nEnter the error tolerance for network testing: ";
        cin>>tolerance;
        cin.ignore(1);
    }
    else
        train_data_file>>input_learn_rate>>hidden_learn_rate>>steepness>>
                         momentum>>max_epochs>>desired_net_error>>tolerance;
}

//this function cleans up all the memory (releases used memory), writes out
//the final weight values, and finally closes all of the files
void clean_up(int argc)
{
    int c;
    neuron *temp_n;
    synapse *temp_s;

    //write out all the weight and bias values for the net to the file
    temp_s=input_s;
    for(c=0;c<no_synapses;c++)
    {
        weights_save_data_file<<temp_s->weight<<" ";
        temp_s++;
    }
    weights_save_data_file<<endl;
    temp_n=input_n;
    for(c=0;c<no_neurons;c++)
    {
        weights_save_data_file<<temp_n->bias<<" ";
        temp_n++;
    }
    //all done, than close all the data files
    train_data_file.close();
    test_data_file.close();
    weights_save_data_file.close();
    weights_original_file.close();
    results_data_file.close();
    if(argc==5)
        weights_load_data_file.close();
    //free up all the used memory
    delete [] input_n;
    delete [] input_s;
}




